# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# --------------------------------------------------------------------------

import sys

from .trace import EventTypes
from .. import utils

logger = utils.get_logger()


class BaseNode:
    def __init__(self):
        self.name = None
        self.start_time = None
        self.end_time = None
        self.type = None
        self.external_id = None  # For consistency check.


class HostNode(BaseNode):
    def __init__(self):
        super(HostNode, self).__init__()
        self.device_duration = 0  # Total time of Kernel, GPU Memcpy, GPU Memset. TODO: parallel multi-stream?


class OperatorNode(HostNode):
    def __init__(self):
        super(OperatorNode, self).__init__()
        self.children = []  # OperatorNode and ProfilerStepNode.
        self.runtimes = []  # RuntimeNode
        self.input_shape = None
        self.self_host_duration = 0
        self.self_device_duration = 0

    def fill_stats(self):
        self.self_host_duration = self.end_time - self.start_time
        for child in self.children:
            self.device_duration += child.device_duration
            self.self_host_duration -= (child.end_time - child.start_time)
        for rt in self.runtimes:
            # From PyTorch 1.8 RC1, cpu_self_time does not include runtime's time.
            # So here we keep consistent with it.
            self.self_host_duration -= (rt.end_time - rt.start_time)
            self.device_duration += rt.device_duration
            self.self_device_duration += rt.device_duration


class ProfilerStepNode(OperatorNode):
    def __init__(self):
        super(ProfilerStepNode, self).__init__()


class RuntimeNode(HostNode):
    def __init__(self):
        super(RuntimeNode, self).__init__()
        # One runtime could trigger more than one kernel, such as cudaLaunchCooperativeKernelMultiDevice.
        self.device_nodes = None

    def fill_stats(self):
        if self.device_nodes is None:
            return
        for device_node in self.device_nodes:
            self.device_duration += device_node.end_time - device_node.start_time


class DeviceNode(BaseNode):
    def __init__(self):
        super(DeviceNode, self).__init__()
        self.op_node = None  # The cpu operator that launched it.


class OperatorAgg:
    def __init__(self):
        self.name = None
        self.input_shape = None  # Optional
        self.calls = 0
        self.host_duration = 0
        self.device_duration = 0
        self.self_host_duration = 0
        self.self_device_duration = 0
        # TODO: Think about adding these avgs to UI.
        self.avg_host_duration = 0
        self.avg_device_duration = 0

    def average(self):
        self.avg_host_duration = self.host_duration / self.calls
        self.avg_device_duration = self.device_duration / self.calls


class KernelAggByNameOp:
    def __init__(self):
        self.name = None
        self.op_name = None
        self.calls = 0
        self.total_duration = 0
        self.avg_duration = 0
        self.min_duration = sys.maxsize
        self.max_duration = 0

    def average(self):
        self.avg_duration = self.total_duration / self.calls


class ModuleParser:
    def __init__(self):
        self.tid2tree = {}
        self.cpp_op_list = []  # For Operator-view.
        self.kernel_list = []  # For Kernel-view.
        self.op_list_groupby_name = []  # For Operator-view.
        self.op_list_groupby_name_input = []  # For Operator-view.
        self.kernel_list_groupby_name_op = {}  # For Kernel-view.

    # host_node_list: list of OperatorNode and ProfilerStepNode.
    # zero_rt_list: list of RuntimeNode with external_id=0.
    def _build_tree(self, host_node_list, zero_rt_list):

        def build_tree_relationship(host_node_list, zero_rt_list):
            node_stack = []
            root_node = OperatorNode()
            root_node.start_time = -sys.maxsize - 1
            root_node.end_time = sys.maxsize
            root_node.runtimes = zero_rt_list  # Give the list of RuntimeNode with external_id=0 to root node.
            node_stack.append(root_node)
            for node in host_node_list:
                while True:  # break loop when the node is inserted.
                    tail_node = node_stack[-1]
                    if node.start_time < tail_node.end_time:
                        if node.end_time <= tail_node.end_time:
                            tail_node.children.append(node)
                            node_stack.append(node)
                        else:
                            logger.error("Error in input data: ranges on the same thread should not intersect!"
                                         "Father:({},{},{}) Child:({},{},{})".format(
                                tail_node.name, tail_node.start_time, tail_node.end_time,
                                node.name, node.start_time, node.end_time
                            ))
                        break
                    else:
                        node_stack.pop()
            root_node.name = "CallTreeRoot"
            root_node.type = EventTypes.PYTHON
            return root_node

        # Merge the consecutive calls to same function into one.
        # Just follow the same pattern in torch/autograd/profiler.py,
        # EventList._remove_dup_nodes
        # TODO: Replace recursive by for loop, in case of too deep callstack.
        def remove_dup_nodes(node):
            if node.type == EventTypes.RUNTIME:
                return
            if len(node.children) == 1:
                child = node.children[0]
                if node.name == child.name and node.type == EventTypes.OPERATOR and child.type == EventTypes.OPERATOR:
                    node.children = child.children
                    node.runtimes = child.runtimes  # Keep consistent with autograd profiler.
                    remove_dup_nodes(node)  # This node may have to merge with child's child.
            for child in node.children:
                remove_dup_nodes(child)

        # TODO: Replace recursive by using a stack, in case of too deep callstack.
        def fill_stats(node):
            if node.type != EventTypes.RUNTIME:
                for child in node.children:
                    fill_stats(child)
                for rt in node.runtimes:
                    fill_stats(rt)
                    if rt.device_nodes is not None:
                        for device_node in rt.device_nodes:
                            device_node.op_node = node

            if node.name == "CallTreeRoot":
                node.start_time = node.end_time = None
                for i in range(len(node.children)):
                    if node.children[i].start_time is not None:
                        node.start_time = node.children[i].start_time
                        break
                for i in range(len(node.children) - 1, -1, -1):
                    if node.children[i].end_time is not None:
                        node.end_time = node.children[i].end_time
                        break

            node.fill_stats()
            if type(node) is OperatorNode and node.type == EventTypes.OPERATOR \
                    and not (node.name.startswith("enumerate(DataLoader)#") and node.name.endswith(".__next__")) \
                    and not node.name.startswith("Optimizer."):
                self.cpp_op_list.append(node)
            if node.type == EventTypes.RUNTIME and node.device_nodes is not None:
                self.kernel_list.extend([n for n in node.device_nodes if n.type == EventTypes.KERNEL])

        root_node = build_tree_relationship(host_node_list, zero_rt_list)
        remove_dup_nodes(root_node)
        fill_stats(root_node)
        return root_node

    def parse_events(self, events):

        def parse_event(event, corrid_to_device, corrid_to_runtime, externalid_to_runtime, tid2list, tid2zero_rt_list):

            def build_node(node, event):
                node.name = event.name
                node.start_time = event.ts
                node.end_time = event.ts + event.duration
                node.type = event.type
                if "external id" in event.args:
                    node.external_id = event.args["external id"]
                elif "External id" in event.args:
                    node.external_id = event.args["External id"]

            corrid = event.args["correlation"] if "correlation" in event.args else None
            input_shape = event.args["Input dims"] if "Input dims" in event.args else None
            tid = event.tid
            if event.type in [EventTypes.KERNEL, EventTypes.MEMCPY, EventTypes.MEMSET]:
                device_node = DeviceNode()
                build_node(device_node, event)
                if corrid in corrid_to_runtime:
                    rt_node = corrid_to_runtime[corrid]  # Don't pop it because it may be used by next kernel.
                    if rt_node.device_nodes is None:
                        rt_node.device_nodes = [device_node]
                    else:
                        rt_node.device_nodes.append(device_node)
                    if rt_node.external_id != device_node.external_id:
                        logger.warning(
                            "Runtime and Device-op have same correlation id but with different external id!"
                        )
                else:
                    if corrid not in corrid_to_device:
                        corrid_to_device[corrid] = [device_node]
                    else:
                        corrid_to_device[corrid].append(device_node)
            elif event.type == EventTypes.RUNTIME:
                rt_node = RuntimeNode()
                build_node(rt_node, event)
                corrid_to_runtime[corrid] = rt_node
                if corrid in corrid_to_device:
                    rt_node.device_nodes = []
                    rt_node.device_nodes.extend(corrid_to_device[corrid])
                    for device_node in corrid_to_device[corrid]:
                        if rt_node.external_id != device_node.external_id:
                            logger.warning(
                                "Runtime and Device-op have same correlation id but with different external id!"
                            )
                if rt_node.external_id in externalid_to_runtime:
                    externalid_to_runtime[rt_node.external_id].append(rt_node)
                else:
                    externalid_to_runtime[rt_node.external_id] = [rt_node]
                # Some runtimes has external_id 0, which will not be correlated to any operator.
                # So get them and attach them to root node.
                if rt_node.external_id == 0:
                    if tid not in tid2zero_rt_list:
                        tid2zero_rt_list[tid] = []
                    tid2zero_rt_list[tid].append(rt_node)
            elif event.type in [EventTypes.PYTHON, EventTypes.OPERATOR, EventTypes.PROFILER_STEP]:
                if event.type == EventTypes.PROFILER_STEP:
                    op_node = ProfilerStepNode()
                else:
                    op_node = OperatorNode()
                build_node(op_node, event)
                op_node.input_shape = input_shape
                if tid not in tid2list:
                    tid2list[tid] = []
                tid2list[tid].append(op_node)

        def parse_ops(cpp_op_list):
            def aggregate(key_to_agg, key, op):
                if key not in key_to_agg:
                    key_to_agg[key] = OperatorAgg()
                agg = key_to_agg[key]
                agg.name = op.name
                agg.input_shape = str(op.input_shape)
                agg.calls += 1
                agg.host_duration += op.end_time - op.start_time
                agg.device_duration += op.device_duration
                agg.self_host_duration += op.self_host_duration
                agg.self_device_duration += op.self_device_duration
                return agg

            name_to_agg = {}
            for op in cpp_op_list:
                agg = aggregate(name_to_agg, op.name, op)
            for _, agg in name_to_agg.items():
                agg.average()
            op_list_groupby_name = list(name_to_agg.values())

            name_input_to_agg = {}
            for op in cpp_op_list:
                name_input = op.name + "###" + str(op.input_shape)
                agg = aggregate(name_input_to_agg, name_input, op)
            for _, agg in name_input_to_agg.items():
                agg.average()
            op_list_groupby_name_input = list(name_input_to_agg.values())

            return op_list_groupby_name, op_list_groupby_name_input

        def parse_kernels(kernel_list):
            name_op_to_agg = {}
            for kernel in kernel_list:
                key = kernel.name + "###" + kernel.op_node.name
                if key not in name_op_to_agg:
                    name_op_to_agg[key] = KernelAggByNameOp()
                agg = name_op_to_agg[key]
                agg.name = kernel.name
                agg.op_name = kernel.op_node.name
                agg.calls += 1
                dur = kernel.end_time - kernel.start_time
                agg.total_duration += dur
                agg.min_duration = min(agg.min_duration, dur)
                agg.max_duration = max(agg.max_duration, dur)
            for _, agg in name_op_to_agg.items():
                agg.average()
            kernel_list_groupby_name_op = list(name_op_to_agg.values())

            return kernel_list_groupby_name_op

        # For OperatorNode and ProfilerStepNode:
        #   Use time interval containing relationship to build father-child correlation,
        #   which is consistent with autograd profiler.
        # For RuntimeNode:
        #   Use external_id to build correlation with its father OperatorNode or ProfilerStepNode.
        #   Because in the case when RuntimeNode has duration 0 and starts at same time as a OperatorNode,
        #   just use interval containing relationship can't tell it is child or brother of the OperatorNode.
        tid2list = {}  # value is a list of OperatorNode and ProfilerStepNode. Do not include RuntimeNode
        tid2zero_rt_list = {}  # value is a list of RuntimeNode with external_id=0. They will be attached to root nodes.
        corrid_to_device = {}  # value is a list of DeviceNode
        corrid_to_runtime = {}  # value is a RuntimeNode
        externalid_to_runtime = {}  # value is a list of RuntimeNode
        for event in events:
            parse_event(event, corrid_to_device, corrid_to_runtime, externalid_to_runtime, tid2list, tid2zero_rt_list)
        # associate CUDA Runtimes with CPU events
        for _, op_list in tid2list.items():
            for op in op_list:
                if op.external_id in externalid_to_runtime:
                    op.runtimes.extend(externalid_to_runtime[op.external_id])
                    del externalid_to_runtime[op.external_id]
        for ext_id in externalid_to_runtime:
            if ext_id != 0:
                logger.warning("{} Runtime with external id {} don't correlate to any operator!".format(
                    len(externalid_to_runtime[ext_id]), ext_id))
        for tid, op_list in tid2list.items():
            zero_rt_list = tid2zero_rt_list[tid] if tid in tid2zero_rt_list else []
            # Note that when 2 start_time are equal, the one with bigger end_time should be ahead of the other.
            op_list.sort(key=lambda x: (x.start_time, -x.end_time))
            root_node = self._build_tree(op_list, zero_rt_list)
            self.tid2tree[tid] = root_node
        self.op_list_groupby_name, self.op_list_groupby_name_input = parse_ops(self.cpp_op_list)
        self.kernel_list_groupby_name_op = parse_kernels(self.kernel_list)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# --------------------------------------------------------------------------

import pandas as pd


class KernelParser:
    def __init__(self):
        self.kernel_stat = None

    def parse_events(self, events):
        events_dict = []
        for event in events:
            events_dict.append(event.to_dict())
        events = events_dict
        events = pd.DataFrame(events)
        events = events.astype({"type": "category", "category": "category", "name": "string"}, copy=False)
        kernels = events[events["category"] == "Kernel"]
        self.kernel_stat = kernels.groupby("name")["duration"].agg(["count", "sum", "mean", "max", "min"]) \
            .sort_values("sum", ascending=False)



import os
import socket
import time
import urllib
import urllib.request
import unittest
from subprocess import Popen


class TestEnd2End(unittest.TestCase):

    def test_tensorboard_end2end(self):
        test_folder = os.path.join(os.path.dirname(os.path.abspath(__file__)),'../samples')
        tb = Popen(['tensorboard', '--logdir='+test_folder])

        run_link = "http://localhost:6006/data/plugin/pytorch_profiler/runs"
        expected_runs = b'["resnet50_num_workers_0", "resnet50_num_workers_4"]'
        host='localhost'
        port=6006

        timeout = 60
        while True:
            try:
                socket.socket(socket.AF_INET, socket.SOCK_STREAM).connect((host, port))
                print('tensorboard start successfully')
                break
            except socket.error:
                time.sleep(2)
                timeout -= 1
                if timeout < 0:
                    tb.kill()
                    raise RuntimeError("tensorboard start timeout")
                continue

        timeout = 60
        while True:
            try:
                response = urllib.request.urlopen(run_link)
                if response.read()==expected_runs:
                    break
                time.sleep(2)
                timeout -= 1
                if timeout<0:
                    tb.kill()
                    raise RuntimeError("Load run timeout")
            except Exception:
                continue

        link_prefix = 'http://localhost:6006/data/plugin/pytorch_profiler/'
        expected_links_format=[]
        expected_links_format.append(link_prefix + 'overview?run={}&worker=worker0&view=Overview')
        expected_links_format.append(link_prefix + 'operation?run={}&worker=worker0&view=Operator&group_by=Operation')
        expected_links_format.append(link_prefix + 'operation/table?run={}&worker=worker0&view=Operator&group_by=Operation')
        expected_links_format.append(link_prefix + 'kernel/table?run={}&worker=worker0&view=Kernel&group_by=Kernel')
        expected_links_format.append(link_prefix + 'kernel?run={}&worker=worker0&view=Kernel&group_by=Kernel')
        links=[]
        for run in ["resnet50_num_workers_0",
                    "resnet50_num_workers_4"]:
            for expected_link in expected_links_format:
                links.append(expected_link.format(run))

        try:
            with open('result_check_file.txt', 'r') as f:
                lines=f.readlines()
                i = 0
                for link in links:
                    response = urllib.request.urlopen(link)
                    self.assertEqual(response.read(), lines[i].strip().encode(encoding="utf-8"))
                    i = i + 1
            self.assertEqual(i, 10)
        finally:
            tb.kill()




# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# --------------------------------------------------------------------------

import sys

from .. import utils
from .trace import EventTypes

logger = utils.get_logger()


def merge_ranges(src_ranges, is_sorted=False):
    merged_ranges = []
    if len(src_ranges) > 0:
        if not is_sorted:
            src_ranges.sort(key=lambda x: x[0])
        src_id = 0
        merged_ranges.append(
            (src_ranges[src_id][0], src_ranges[src_id][1]))
        for src_id in range(1, len(src_ranges)):
            dst_id = len(merged_ranges) - 1
            if src_ranges[src_id][1] > merged_ranges[dst_id][1]:
                if src_ranges[src_id][0] <= merged_ranges[dst_id][1]:
                    merged_ranges[dst_id] = (merged_ranges[dst_id][0], src_ranges[src_id][1])
                else:
                    merged_ranges.append(
                        (src_ranges[src_id][0], src_ranges[src_id][1]))
    return merged_ranges


def subtract_ranges_lists(range_list1, range_list2):
    range_list_dst = []
    if len(range_list1) == 0:
        return range_list_dst
    if len(range_list2) == 0:
        range_list_dst = list(range_list1)
        return range_list_dst
    r1 = range_list1[0]
    r2 = range_list2[0]
    i1 = i2 = 0
    while i1 < len(range_list1):
        if i2 == len(range_list2):
            range_list_dst.append(r1)
            r1, i1 = pop_list(range_list1, i1)
        elif r2[1] <= r1[0]:
            r2, i2 = pop_list(range_list2, i2)
        elif r2[0] <= r1[0] and r2[1] < r1[1]:
            r1 = (r2[1], r1[1])
            r2, i2 = pop_list(range_list2, i2)
        elif r2[0] <= r1[0]:
            assert (r2[1] >= r1[1])
            r2 = (r1[1], r2[1])
            r1, i1 = pop_list(range_list1, i1)
        elif r2[0] < r1[1]:
            assert (r2[0] > r1[0])
            range_list_dst.append((r1[0], r2[0]))
            r1 = (r2[0], r1[1])
        else:
            assert (r2[0] >= r1[1])
            range_list_dst.append(r1)
            r1, i1 = pop_list(range_list1, i1)
    return range_list_dst


def intersection_ranges_lists(range_list1, range_list2):
    range_list_dst = []
    if len(range_list1) == 0 or len(range_list2) == 0:
        return range_list_dst
    r1 = range_list1[0]
    r2 = range_list2[0]
    i1 = i2 = 0
    while i1 < len(range_list1):
        if i2 == len(range_list2):
            break
        elif r2[1] <= r1[0]:
            r2, i2 = pop_list(range_list2, i2)
        elif r2[0] <= r1[0] and r2[1] < r1[1]:
            assert (r2[1] > r1[0])
            range_list_dst.append((r1[0], r2[1]))
            r1 = (r2[1], r1[1])
            r2, i2 = pop_list(range_list2, i2)
        elif r2[0] <= r1[0]:
            assert (r2[1] >= r1[1])
            range_list_dst.append(r1)
            r2 = (r1[1], r2[1])
            r1, i1 = pop_list(range_list1, i1)
        elif r2[1] < r1[1]:
            assert (r2[0] > r1[0])
            range_list_dst.append(r2)
            r1 = (r2[1], r1[1])
            r2, i2 = pop_list(range_list2, i2)
        elif r2[0] < r1[1]:
            assert (r2[1] >= r1[1])
            range_list_dst.append((r2[0], r1[1]))
            r2 = (r1[1], r2[1])
            r1, i1 = pop_list(range_list1, i1)
        else:
            assert (r2[0] >= r1[1])
            r1, i1 = pop_list(range_list1, i1)
    return range_list_dst


def get_ranges_sum(ranges):
    sum = 0
    for range in ranges:
        sum += (range[1] - range[0])
    return sum


def pop_list(range_list, index):
    next_index = index + 1
    if next_index >= len(range_list):
        return None, len(range_list)
    next_item = range_list[next_index]
    return next_item, next_index


class OverallParser(object):
    class Costs:
        def __init__(self):
            self.step_total_cost = 0
            self.kernel_cost = 0
            self.memcpy_cost = 0
            self.memset_cost = 0
            self.runtime_cost = 0
            self.dataloader_cost = 0
            self.cpuop_cost = 0
            self.other_cost = 0

        def calculate_costs(self, statistics, step):
            self.step_total_cost = step[1] - step[0]
            self.kernel_cost = get_ranges_sum(statistics.kernel_cost_ranges)
            self.memcpy_cost = get_ranges_sum(statistics.memcpy_cost_ranges)
            self.memset_cost = get_ranges_sum(statistics.memset_cost_ranges)
            self.runtime_cost = get_ranges_sum(statistics.runtime_cost_ranges)
            self.dataloader_cost = get_ranges_sum(statistics.dataloader_cost_ranges)
            self.cpuop_cost = get_ranges_sum(statistics.cpuop_cost_ranges)
            self.other_cost = get_ranges_sum(statistics.other_cost_ranges)

    class Statistics:
        def __init__(self):
            self.kernel_cost_ranges = []
            self.memcpy_cost_ranges = []
            self.memset_cost_ranges = []
            self.runtime_cost_ranges = []
            self.dataloader_cost_ranges = []
            self.cpuop_cost_ranges = []
            self.other_cost_ranges = []

        def intersection_with_step(self, step):
            result = OverallParser.Statistics()
            step = [step]
            result.kernel_cost_ranges = intersection_ranges_lists(step, self.kernel_cost_ranges)
            result.memcpy_cost_ranges = intersection_ranges_lists(step, self.memcpy_cost_ranges)
            result.memset_cost_ranges = intersection_ranges_lists(step, self.memset_cost_ranges)
            result.runtime_cost_ranges = intersection_ranges_lists(step, self.runtime_cost_ranges)
            result.dataloader_cost_ranges = intersection_ranges_lists(step, self.dataloader_cost_ranges)
            result.cpuop_cost_ranges = intersection_ranges_lists(step, self.cpuop_cost_ranges)
            result.other_cost_ranges = intersection_ranges_lists(step, self.other_cost_ranges)
            return result

    def __init__(self):
        self.kernel_ranges = []
        self.memcpy_ranges = []
        self.memset_ranges = []
        self.runtime_ranges = []
        self.dataloader_ranges = []
        self.cpuop_ranges = []
        self.steps = []
        self.steps_names = []
        self.is_gpu_used = False
        self.min_ts = sys.maxsize
        self.max_ts = -sys.maxsize - 1
        self.steps_costs = []
        self.avg_costs = OverallParser.Costs()

    def parse_events(self, events):
        logger.debug("Overall, parse events")
        for event in events:
            self.parse_event(event)

        self.kernel_ranges = merge_ranges(self.kernel_ranges)
        self.memcpy_ranges = merge_ranges(self.memcpy_ranges)
        self.memset_ranges = merge_ranges(self.memset_ranges)
        self.runtime_ranges = merge_ranges(self.runtime_ranges)
        self.dataloader_ranges = merge_ranges(self.dataloader_ranges)
        self.cpuop_ranges = merge_ranges(self.cpuop_ranges)
        if len(self.steps) == 0:
            self.steps.append((self.min_ts, self.max_ts))
            self.steps_names.append("0")
        merged_steps = list(self.steps)
        merged_steps = merge_ranges(merged_steps)

        logger.debug("Overall, statistics")
        global_stats = OverallParser.Statistics()
        global_stats.kernel_cost_ranges = self.kernel_ranges
        slots = subtract_ranges_lists(merged_steps, self.kernel_ranges)
        global_stats.memcpy_cost_ranges = intersection_ranges_lists(slots, self.memcpy_ranges)
        slots = subtract_ranges_lists(slots, global_stats.memcpy_cost_ranges)
        global_stats.memset_cost_ranges = intersection_ranges_lists(slots, self.memset_ranges)
        slots = subtract_ranges_lists(slots, global_stats.memset_cost_ranges)
        global_stats.runtime_cost_ranges = intersection_ranges_lists(slots, self.runtime_ranges)
        slots = subtract_ranges_lists(slots, global_stats.runtime_cost_ranges)
        global_stats.dataloader_cost_ranges = intersection_ranges_lists(slots, self.dataloader_ranges)
        slots = subtract_ranges_lists(slots, global_stats.dataloader_cost_ranges)
        global_stats.cpuop_cost_ranges = intersection_ranges_lists(slots, self.cpuop_ranges)
        slots = subtract_ranges_lists(slots, global_stats.cpuop_cost_ranges)
        global_stats.other_cost_ranges = slots

        logger.debug("Overall, aggregation")
        valid_steps = len(self.steps)
        for i in range(valid_steps):
            steps_stat = global_stats.intersection_with_step(self.steps[i])
            self.steps_costs.append(OverallParser.Costs())
            self.steps_costs[i].calculate_costs(steps_stat, self.steps[i])
            self.avg_costs.step_total_cost += self.steps_costs[i].step_total_cost
            self.avg_costs.kernel_cost += self.steps_costs[i].kernel_cost
            self.avg_costs.memcpy_cost += self.steps_costs[i].memcpy_cost
            self.avg_costs.memset_cost += self.steps_costs[i].memset_cost
            self.avg_costs.runtime_cost += self.steps_costs[i].runtime_cost
            self.avg_costs.dataloader_cost += self.steps_costs[i].dataloader_cost
            self.avg_costs.cpuop_cost += self.steps_costs[i].cpuop_cost
            self.avg_costs.other_cost += self.steps_costs[i].other_cost

        self.avg_costs.step_total_cost /= valid_steps
        self.avg_costs.kernel_cost /= valid_steps
        self.avg_costs.memcpy_cost /= valid_steps
        self.avg_costs.memset_cost /= valid_steps
        self.avg_costs.runtime_cost /= valid_steps
        self.avg_costs.dataloader_cost /= valid_steps
        self.avg_costs.cpuop_cost /= valid_steps
        self.avg_costs.other_cost /= valid_steps

    def parse_event(self, event):
        ts = event.ts
        dur = event.duration
        evt_type = event.type
        if evt_type == EventTypes.KERNEL:
            self.kernel_ranges.append((ts, ts + dur))
        elif evt_type == EventTypes.MEMCPY:
            self.memcpy_ranges.append((ts, ts + dur))
        elif evt_type == EventTypes.MEMSET:
            self.memset_ranges.append((ts, ts + dur))
        elif evt_type == EventTypes.RUNTIME:
            self.runtime_ranges.append((ts, ts + dur))
        elif evt_type == EventTypes.OPERATOR and event.name.startswith("enumerate(DataLoader)#") \
                and event.name.endswith(".__next__"):
            self.dataloader_ranges.append((ts, ts + dur))
        elif event.type == EventTypes.PROFILER_STEP:
            self.steps.append((ts, ts + dur))
            self.steps_names.append(str(event.step))
        elif evt_type in [EventTypes.PYTHON, EventTypes.OPERATOR]:
            self.cpuop_ranges.append((ts, ts + dur))

        if evt_type == EventTypes.RUNTIME:
            self.is_gpu_used = True

        if ts < self.min_ts:
            self.min_ts = ts
        if ts + dur > self.max_ts:
            self.max_ts = ts + dur



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# --------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import setuptools

def get_version():
    with open("version.txt", encoding="utf-8") as f:
        version = f.read().strip()

    if os.getenv('TORCH_TB_PROFILER_BUILD_VERSION'):
        version = os.getenv('TORCH_TB_PROFILER_BUILD_VERSION')
    return version

def get_readme():
    with open("README.md", encoding="utf-8") as f:
        return f.read()

INSTALL_REQUIRED = [
    "pandas",
    "tensorboard >= 1.15, !=2.1.0"
]

TESTS_REQUIRED = INSTALL_REQUIRED + [
    "torch >= 1.8",
    "torchvision >= 0.8"
]

setuptools.setup(
    name="torch_tb_profiler",
    version=get_version(),
    description="PyTorch Profiler TensorBoard Plugin",
    long_description=get_readme(),
    long_description_content_type='text/markdown',
    url="https://github.com/pytorch/kineto/tree/master/tb_plugin",
    author="Microsoft Inc.",
    packages=setuptools.find_packages(),
    package_data={
        "torch_tb_profiler": ["static/**"],
    },
    entry_points={
        "tensorboard_plugins": [
            "torch_profiler = torch_tb_profiler.plugin:TorchProfilerPlugin",
        ],
    },
    python_requires=">= 2.7, != 3.0.*, != 3.1.*",
    install_requires=INSTALL_REQUIRED,
    tests_require=TESTS_REQUIRED,
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    license='BSD-3',
    keywords='pytorch tensorboard profile plugin',
)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# --------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from collections import namedtuple

PLUGIN_NAME = "pytorch_profiler"

TRACE_FILE_SUFFIX = ".pt.trace.json"
TRACE_GZIP_FILE_SUFFIX = ".pt.trace.json.gz"

MONITOR_RUN_REFRESH_INTERNAL_IN_SECONDS = 10

View = namedtuple("View", "id, name, display_name")
OVERALL_VIEW = View(1, "overall", "Overview")
OP_VIEW = View(2, "operator", "Operator")
KERNEL_VIEW = View(3, "kernel", "Kernel")
TRACE_VIEW = View(4, "trace", "Trace")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# --------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from collections import OrderedDict


class Run(object):
    """ A profiler run. For visualization purpose only.
    May contain profiling results from multiple workers. E.g. distributed scenario.
    """

    def __init__(self, name, run_dir):
        self.name = name
        self.run_dir = run_dir
        self.profiles = OrderedDict()

    @property
    def workers(self):
        return list(self.profiles.keys())

    @property
    def views(self):
        profile = self.get_profile()
        if profile is None:
            return None
        return profile.views

    def add_profile(self, profile):
        self.profiles[profile.worker] = profile

    def get_profile(self, worker=None):
        if len(self.profiles) == 0:
            return None
        if not worker:
            return next(iter(self.profiles.values()))
        return self.profiles.get(worker, None)


class RunProfile(object):
    """ Cooked profiling result for a worker. For visualization purpose only.
    """

    def __init__(self, worker):
        self.worker = worker
        self.views = []
        self.is_gpu_used = False
        self.overview = None
        self.operation_pie_by_name = None
        self.operation_table_by_name = None
        self.operation_pie_by_name_input = None
        self.operation_table_by_name_input = None
        self.kernel_op_table = None
        self.kernel_pie = None
        self.kernel_table = None
        self.trace_file_path = None
